from pathlib import Path
from unittest.mock import MagicMock, patch

from mkdocs_zettelkasten.plugin.services.zettel_service import ZettelService

VALID_MD = "---\nid: 1\ndate: 2024-01-01\n---\n# Title\nBody\n"


class TestZettelService:
    def test_configure_stores_config(self) -> None:
        svc = ZettelService()
        cfg = {"id_key": "myid"}
        svc.configure(cfg)
        assert svc.zettel_config == cfg

    def test_process_files(self, tmp_path: Path) -> None:
        fp = tmp_path / "note.md"
        fp.write_text(VALID_MD)

        f = MagicMock()
        f.src_path = "note.md"
        f.abs_src_path = str(fp)
        f.is_documentation_page.return_value = True

        files = MagicMock()
        files.__iter__ = lambda self: iter([f])

        config = MagicMock()
        config.__getitem__ = lambda self, key: {"docs_dir": str(tmp_path)}[key]

        svc = ZettelService()
        with patch(
            "mkdocs_zettelkasten.plugin.entities.zettel.GitUtil.is_tracked",
            return_value=False,
        ):
            svc.process_files(files, config)

        assert len(svc.get_zettels()) == 1

    def test_add_zettel_to_page(self, tmp_path: Path) -> None:
        fp = tmp_path / "note.md"
        fp.write_text(VALID_MD)

        f = MagicMock()
        f.src_path = "note.md"
        f.abs_src_path = str(fp)
        f.is_documentation_page.return_value = True

        files = MagicMock()
        files.__iter__ = lambda self: iter([f])

        config = MagicMock()
        config.__getitem__ = lambda self, key: {"docs_dir": str(tmp_path)}[key]

        svc = ZettelService()
        with patch(
            "mkdocs_zettelkasten.plugin.entities.zettel.GitUtil.is_tracked",
            return_value=False,
        ):
            svc.process_files(files, config)

        page = MagicMock()
        page.file.abs_src_path = str(fp)
        page.meta = {}

        result = svc.add_zettel_to_page(page)
        assert result.meta["is_zettel"] is True
        assert result.meta["zettel"].id == 1

    def test_add_zettel_non_zettel_page(self) -> None:
        svc = ZettelService()
        page = MagicMock()
        page.file.abs_src_path = None
        page.meta = {}

        result = svc.add_zettel_to_page(page)
        assert result.meta["is_zettel"] is False



"""Shared frontmatter parsing — single source of truth for YAML header splitting."""

from __future__ import annotations

DIVIDER = "---"
DIVIDER_COUNT = 2


def parse_frontmatter(content: str) -> tuple[str, str]:
    """Split content into (header_yaml_text, body_text) at YAML frontmatter boundary.

    Returns ("", content) if no valid frontmatter found.
    """
    lines = content.splitlines(keepends=True)
    divider_count = 0
    header_start = 0

    for i, line in enumerate(lines):
        if line.strip() == DIVIDER:
            divider_count += 1
            if divider_count == 1:
                header_start = i + 1
            elif divider_count == DIVIDER_COUNT:
                header = "".join(lines[header_start:i])
                body = "".join(lines[i + 1 :])
                return (header, body)

    return ("", content)



from pathlib import Path
from unittest.mock import MagicMock

from mkdocs_zettelkasten.plugin.services.backlink_processor import BacklinkProcessor
from mkdocs_zettelkasten.plugin.services.zettel_store import ZettelStore


def _make_zettel(zettel_id: int, path: str, links: list[str]) -> MagicMock:
    z = MagicMock()
    z.id = zettel_id
    z.path = Path(path)
    z.links = links
    z.__hash__ = lambda self: hash(zettel_id)
    z.__eq__ = lambda self, other: getattr(other, "id", None) == zettel_id
    return z


class TestBacklinkProcessor:
    def test_empty_store(self) -> None:
        store = ZettelStore()
        result = BacklinkProcessor.process(store)
        assert result == {}

    def test_creates_backlinks(self) -> None:
        z1 = _make_zettel(1, "/docs/a.md", ["b.md"])
        z2 = _make_zettel(2, "/docs/b.md", [])

        store = ZettelStore([z1, z2])
        result = BacklinkProcessor.process(store)

        assert "b.md" in result
        assert z1 in result["b.md"]

    def test_normalizes_links_without_suffix(self) -> None:
        z1 = _make_zettel(1, "/docs/a.md", ["b"])
        z2 = _make_zettel(2, "/docs/b.md", [])

        store = ZettelStore([z1, z2])
        result = BacklinkProcessor.process(store)

        assert "b.md" in result

    def test_normalize_links_appends_suffix(self) -> None:
        result = BacklinkProcessor.normalize_links(["a", "b.md"])
        assert result == {"a.md", "b.md"}

    def test_normalize_links_custom_suffix(self) -> None:
        result = BacklinkProcessor.normalize_links(["a", "b.txt"], file_suffix=".txt")
        assert result == {"a.txt", "b.txt"}

    def test_normalize_links_deduplicates(self) -> None:
        result = BacklinkProcessor.normalize_links(["a", "a.md"])
        assert result == {"a.md"}

    def test_ignores_unresolved_links(self) -> None:
        z1 = _make_zettel(1, "/docs/a.md", ["nonexistent"])
        store = ZettelStore([z1])
        result = BacklinkProcessor.process(store)
        assert result == {}



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from mkdocs.config.defaults import MkDocsConfig
    from mkdocs.structure.files import Files
    from mkdocs.structure.pages import Page

    from mkdocs_zettelkasten.plugin.entities.zettel import Zettel

import logging

from .backlink_processor import BacklinkProcessor
from .zettel_parser import ZettelParser
from .zettel_store import ZettelStore

logger = logging.getLogger(
    __name__.replace("mkdocs_zettelkasten.plugin.", "mkdocs.plugins.zettelkasten.")
)


class ZettelService:
    """Orchestrates zettel processing pipeline."""

    def __init__(self) -> None:
        self.store = ZettelStore()
        self.backlinks: dict[str, list[Zettel]] = {}
        self.invalid_files: list = []
        self.zettel_config: dict[str, Any] = {}
        self.file_suffix: str = ".md"

    def configure(self, zettel_config: dict[str, Any]) -> None:
        self.zettel_config = zettel_config
        self.file_suffix = zettel_config.get("file_suffix", ".md")

    def process_files(self, files: Files, config: MkDocsConfig) -> None:
        """Main processing pipeline."""
        docs_dir = config["docs_dir"]
        logger.info("Scanning `%s` for zettels", docs_dir)
        valid_zettels, self.invalid_files = ZettelParser.parse_files(
            files, self.zettel_config
        )
        logger.info("Found %s zettels in `%s`", len(valid_zettels), docs_dir)
        self.store.update(valid_zettels)
        self.backlinks = BacklinkProcessor.process(
            self.store, file_suffix=self.file_suffix
        )

    def add_zettel_to_page(self, page: Page) -> Page:
        enriched_page = page

        if not enriched_page.file.abs_src_path:
            enriched_page.meta["is_zettel"] = False
            return enriched_page

        zettel_for_page = self.get_zettel_by_path(Path(enriched_page.file.abs_src_path))

        if zettel_for_page:
            enriched_page.meta["zettel"] = zettel_for_page
            enriched_page.meta["is_zettel"] = True
            logger.debug("Added zettel to %s's metadata.", page.file.src_path)
        else:
            enriched_page.meta["is_zettel"] = False
            logger.debug(
                "Skipped adding zettel to %s's metadata, as no zettel was found.",
                page.file.src_path,
            )

        return enriched_page

    def get_zettels(self) -> list[Zettel]:
        return list(self.store.zettels)

    def get_zettel_by_path(self, path: Path) -> Zettel | None:
        """Delegate to store's lookup mechanism."""
        return self.store.get_by_path(path)

    def get_zettel_by_partial_path(self, partial_path: str) -> Zettel | None:
        """Delegate to store's lookup mechanism."""
        return self.store.get_by_partial_path(partial_path)



from __future__ import annotations

import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from re import Match

    from mkdocs.config.defaults import MkDocsConfig
    from mkdocs.structure.files import Files
    from mkdocs.structure.pages import Page

    from mkdocs_zettelkasten.plugin.entities.zettel import Zettel

import logging

from mkdocs_zettelkasten.plugin.utils.patterns import MD_LINK, WIKI_LINK

logger = logging.getLogger(
    __name__.replace("mkdocs_zettelkasten.plugin.", "mkdocs.plugins.zettelkasten.")
)


def adapt_page_links_to_zettels(
    markdown: str,
    page: Page,
    config: MkDocsConfig,
    files: Files,
    zettel_lookup: Callable[[str], Zettel | None],
    file_suffix: str = ".md",
) -> str:
    """Adapt links in the markdown to point to zettels."""

    def process_match(m: Match) -> str:
        url = m.groupdict().get("url", "")
        title = m.groupdict().get("title", url)

        for f in files:
            url_with_suffix = (
                url + file_suffix if not url.endswith(file_suffix) else url
            )

            if url_with_suffix in f.src_path:
                if f.page and (
                    title == url_with_suffix or title + file_suffix == url_with_suffix
                ):
                    target_zettel = zettel_lookup(url_with_suffix)
                    title = target_zettel.title if target_zettel else url
                new_url = config["site_url"] + f.url
                logger.debug(
                    "Transformed link %s to [%s](%s) in %s",
                    m.group(),
                    title,
                    new_url,
                    page.file.src_path,
                )

                return f"[{title}]({new_url})"

        if f"[{title}]({url})" != m.group():
            logger.debug(
                "Transformed link %s to [%s](%s) in %s",
                m.group(),
                title,
                url,
                page.file.src_path,
            )

        return f"[{title}]({url})"

    # Split markdown into alternating segments (non-code/code)
    code_block_pattern = re.compile(r"```", re.DOTALL)
    parts = code_block_pattern.split(markdown)

    processed_parts = []
    for i, original_part in enumerate(parts):
        processed_part = original_part
        if i % 2 == 0:  # Process even-indexed non-code segments
            processed_part = WIKI_LINK.sub(process_match, processed_part)
            processed_part = MD_LINK.sub(process_match, processed_part)
        processed_parts.append(processed_part)

    return "```".join(processed_parts)



from playwright.sync_api import Page


def test_katex_css_loaded(page: Page, math_site: str) -> None:
    page.goto(f"{math_site}/20990201000003/")
    link = page.locator("link[href*='katex']")
    assert link.count() == 1


def test_katex_js_loaded(page: Page, math_site: str) -> None:
    page.goto(f"{math_site}/20990201000003/")
    script = page.locator("script[src*='katex']")
    assert script.count() == 1


def test_inline_math_rendered(page: Page, math_site: str) -> None:
    page.goto(f"{math_site}/20990201000003/")
    page.locator("span.arithmatex .katex").first.wait_for(
        state="attached", timeout=5000
    )
    inline = page.locator("span.arithmatex .katex")
    assert inline.count() > 0


def test_display_math_rendered(page: Page, math_site: str) -> None:
    page.goto(f"{math_site}/20990201000003/")
    page.locator("div.arithmatex .katex-display").first.wait_for(
        state="attached", timeout=5000
    )
    display = page.locator("div.arithmatex .katex-display")
    assert display.count() > 0


def test_katex_not_loaded_when_disabled(page: Page, default_site: str) -> None:
    page.goto(default_site)
    link = page.locator("link[href*='katex']")
    assert link.count() == 0
    script = page.locator("script[src*='katex']")
    assert script.count() == 0



from unittest.mock import MagicMock, patch

from mkdocs_zettelkasten.plugin.services.page_transformer import PageTransformer

MODULE = "mkdocs_zettelkasten.plugin.services.page_transformer"


def _make_transform_fixtures():
    transformer = PageTransformer()
    page = MagicMock()
    page.file.src_path = "test.md"
    page.meta = {"is_zettel": False}
    config = MagicMock()
    config.__getitem__ = lambda self, key: {"site_url": "https://example.com/"}.get(
        key, MagicMock()
    )
    config.get = lambda key, default=None: {"extra": {}}.get(key, default)
    files = MagicMock()
    zettel_service = MagicMock()
    zettel_service.add_zettel_to_page.return_value = page
    zettel_service.get_zettels.return_value = []
    zettel_service.file_suffix = ".md"
    return transformer, page, config, files, zettel_service


class TestPageTransformer:
    def test_transform_calls_all_adapters(self) -> None:
        transformer, page, config, files, zettel_service = _make_transform_fixtures()

        with (
            patch(f"{MODULE}.adapt_page_title", return_value="md1") as mock_title,
            patch(f"{MODULE}.adapt_transclusion", return_value="md1t"),
            patch(
                f"{MODULE}.adapt_page_links_to_zettels", return_value="md2"
            ) as mock_links,
            patch(f"{MODULE}.get_page_ref", return_value=("md3", None)) as mock_ref,
            patch(
                f"{MODULE}.get_prev_next_page", return_value=(None, None)
            ) as mock_nav,
            patch(f"{MODULE}.adapt_backlinks_to_page") as mock_backlinks,
        ):
            result = transformer.transform(
                "original", page, config, files, zettel_service
            )

        assert result == "md3"
        mock_title.assert_called_once()
        mock_links.assert_called_once()
        mock_ref.assert_called_once()
        mock_nav.assert_called_once()
        mock_backlinks.assert_called_once()

    def test_transform_chains_adapters_in_order(self) -> None:
        """Each adapter receives the output of the previous one."""
        transformer, page, config, files, zettel_service = _make_transform_fixtures()

        with (
            patch(f"{MODULE}.adapt_page_title", return_value="after_title") as m_title,
            patch(
                f"{MODULE}.adapt_transclusion", return_value="after_transclusion"
            ) as m_trans,
            patch(
                f"{MODULE}.adapt_page_links_to_zettels", return_value="after_links"
            ) as m_links,
            patch(f"{MODULE}.get_page_ref", return_value=("after_ref", None)) as m_ref,
            patch(f"{MODULE}.get_prev_next_page", return_value=(None, None)),
            patch(f"{MODULE}.adapt_backlinks_to_page"),
        ):
            result = transformer.transform(
                "original", page, config, files, zettel_service
            )

        # title gets the raw markdown
        assert m_title.call_args[0][0] == "original"
        # transclusion gets title output
        assert m_trans.call_args[0][0] == "after_title"
        # links gets transclusion output
        assert m_links.call_args[0][0] == "after_transclusion"
        # ref gets links output
        assert m_ref.call_args[0][0] == "after_links"
        # final result is ref output
        assert result == "after_ref"



from __future__ import annotations

import logging
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from mkdocs.config.defaults import MkDocsConfig
    from mkdocs.structure.files import Files

from mkdocs_zettelkasten.plugin.utils.jinja_utils import create_jinja_environment

logger = logging.getLogger(
    __name__.replace("mkdocs_zettelkasten.plugin.", "mkdocs.plugins.zettelkasten.")
)


@dataclass
class ValidationIssue:
    path: str
    check: str  # invalid_file, orphan, broken_link
    severity: str  # error, warning, info
    message: str


class ValidationService:
    def __init__(self) -> None:
        self.issues: dict[str, list[ValidationIssue]] = defaultdict(list)
        self.output_folder: Path = Path(".build")
        self.output_filename: Path = Path("validation.md")
        self.file_suffix: str = ".md"

    def configure(self, config: MkDocsConfig, file_suffix: str = ".md") -> None:
        self.output_folder = Path(config.get("tags_folder", ".build"))
        if not self.output_folder.is_absolute():
            self.output_folder = Path(config["docs_dir"]).parent / self.output_folder
        self.file_suffix = file_suffix
        if not self.output_folder.exists():
            self.output_folder.mkdir(parents=True)

    def validate(self, zettel_service: Any, files: Files, config: MkDocsConfig) -> None:
        from .backlink_processor import BacklinkProcessor

        self.issues.clear()
        self._check_invalid_files(zettel_service)
        self._check_orphans(zettel_service)
        self._check_broken_links(zettel_service, BacklinkProcessor)
        self._generate_report()
        self._add_to_build(files, config)

        logger.info("Validation complete: %d issues found.", self.total_issues())

    def total_issues(self) -> int:
        return sum(len(v) for v in self.issues.values())

    def get_issues(self, rel_path: str) -> list[ValidationIssue]:
        return self.issues.get(rel_path, [])

    def _check_invalid_files(self, zettel_service: Any) -> None:
        for f in zettel_service.invalid_files:
            self.issues[f.src_path].append(
                ValidationIssue(
                    path=f.src_path,
                    check="invalid_file",
                    severity="error",
                    message="Invalid zettel (missing ID or malformed YAML)",
                )
            )

    def _check_orphans(self, zettel_service: Any) -> None:
        targeted_ids: set[int] = set()
        for link_path in zettel_service.backlinks:
            target = zettel_service.store.get_by_partial_path(
                link_path, self.file_suffix
            )
            if target:
                targeted_ids.add(target.id)

        for zettel in zettel_service.get_zettels():
            if zettel.id not in targeted_ids:
                self.issues[zettel.rel_path].append(
                    ValidationIssue(
                        path=zettel.rel_path,
                        check="orphan",
                        severity="info",
                        message="No backlinks from other zettels",
                    )
                )

    def _check_broken_links(
        self, zettel_service: Any, backlink_processor_cls: type
    ) -> None:
        for zettel in zettel_service.get_zettels():
            internal_links = [
                link
                for link in zettel.links
                if not link.startswith(("http://", "https://", "#", "mailto:"))
            ]
            for link in backlink_processor_cls.normalize_links(
                internal_links, self.file_suffix
            ):
                target = zettel_service.store.get_by_partial_path(
                    link, self.file_suffix
                )
                if not target:
                    logger.warning("Broken link in %s: %s", zettel.rel_path, link)
                    self.issues[zettel.rel_path].append(
                        ValidationIssue(
                            path=zettel.rel_path,
                            check="broken_link",
                            severity="warning",
                            message=f"Broken link: {link}",
                        )
                    )

    def _generate_report(self) -> None:
        env = create_jinja_environment(None)
        template = env.get_template("validation.md.j2")

        all_issues: list[ValidationIssue] = []
        for path_issues in self.issues.values():
            all_issues.extend(path_issues)

        content = template.render(
            errors=[i for i in all_issues if i.severity == "error"],
            warnings=[i for i in all_issues if i.severity == "warning"],
            info=[i for i in all_issues if i.severity == "info"],
            total=len(all_issues),
        )

        output_path = self.output_folder / self.output_filename
        output_path.write_text(content, encoding="utf-8")

    def _add_to_build(self, files: Files, config: MkDocsConfig) -> None:
        from mkdocs.structure.files import File

        new_file = File(
            path=str(self.output_filename),
            src_dir=str(self.output_folder),
            dest_dir=config["site_dir"],
            use_directory_urls=False,
        )
        files.append(new_file)
        logger.info("Validation report added as %s", new_file.src_path)



from pathlib import Path
from unittest.mock import MagicMock

from mkdocs_zettelkasten.plugin.services.validation_service import (
    ValidationService,
)


class TestValidationService:
    def _make_zettel(
        self, zettel_id: int, rel_path: str, links: list[str] | None = None
    ):
        z = MagicMock()
        z.id = zettel_id
        z.rel_path = rel_path
        z.links = links or []
        return z

    def _make_zettel_service(self, zettels, backlinks=None, invalid_files=None):
        svc = MagicMock()
        svc.get_zettels.return_value = zettels
        svc.store.get_by_partial_path = self._make_partial_path_lookup(zettels)
        svc.backlinks = backlinks or {}
        svc.invalid_files = invalid_files or []
        return svc

    def _make_partial_path_lookup(self, zettels):
        def lookup(partial, _file_suffix=".md"):
            for z in zettels:
                if partial in z.rel_path:
                    return z
            return None

        return lookup

    def test_no_issues_when_all_linked(self, tmp_path: Path) -> None:
        z1 = self._make_zettel(1, "a.md", links=["b"])
        z2 = self._make_zettel(2, "b.md", links=["a"])

        svc = self._make_zettel_service(
            [z1, z2], backlinks={"a.md": [z2], "b.md": [z1]}
        )

        vs = ValidationService()
        vs.output_folder = tmp_path
        config = MagicMock()
        config.__getitem__ = lambda self, k: str(tmp_path) if k == "site_dir" else ""
        files = MagicMock()

        vs.validate(svc, files, config)

        assert vs.get_issues("a.md") == []
        assert vs.get_issues("b.md") == []

    def test_orphan_detected(self, tmp_path: Path) -> None:
        z1 = self._make_zettel(1, "a.md", links=["b"])
        z2 = self._make_zettel(2, "b.md", links=[])

        svc = self._make_zettel_service([z1, z2], backlinks={"b.md": [z1]})

        vs = ValidationService()
        vs.output_folder = tmp_path
        config = MagicMock()
        config.__getitem__ = lambda self, k: str(tmp_path) if k == "site_dir" else ""
        files = MagicMock()

        vs.validate(svc, files, config)

        a_issues = vs.get_issues("a.md")
        assert len(a_issues) == 1
        assert a_issues[0].check == "orphan"
        assert vs.get_issues("b.md") == []

    def test_broken_link_detected(self, tmp_path: Path) -> None:
        z1 = self._make_zettel(1, "a.md", links=["nonexistent"])

        svc = self._make_zettel_service([z1], backlinks={})

        vs = ValidationService()
        vs.output_folder = tmp_path
        config = MagicMock()
        config.__getitem__ = lambda self, k: str(tmp_path) if k == "site_dir" else ""
        files = MagicMock()

        vs.validate(svc, files, config)

        issues = vs.get_issues("a.md")
        broken = [i for i in issues if i.check == "broken_link"]
        assert len(broken) == 1
        assert "nonexistent" in broken[0].message

    def test_invalid_file_reported(self, tmp_path: Path) -> None:
        invalid_f = MagicMock()
        invalid_f.src_path = "bad.md"

        svc = self._make_zettel_service([], invalid_files=[invalid_f])

        vs = ValidationService()
        vs.output_folder = tmp_path
        config = MagicMock()
        config.__getitem__ = lambda self, k: str(tmp_path) if k == "site_dir" else ""
        files = MagicMock()

        vs.validate(svc, files, config)

        issues = vs.get_issues("bad.md")
        assert len(issues) == 1
        assert issues[0].check == "invalid_file"
        assert issues[0].severity == "error"

    def test_external_links_not_flagged(self, tmp_path: Path) -> None:
        z1 = self._make_zettel(
            1, "a.md", links=["https://example.com", "#anchor", "mailto:x@y.com"]
        )

        svc = self._make_zettel_service([z1], backlinks={})

        vs = ValidationService()
        vs.output_folder = tmp_path
        config = MagicMock()
        config.__getitem__ = lambda self, k: str(tmp_path) if k == "site_dir" else ""
        files = MagicMock()

        vs.validate(svc, files, config)

        broken = [i for i in vs.get_issues("a.md") if i.check == "broken_link"]
        assert len(broken) == 0

    def test_report_file_generated(self, tmp_path: Path) -> None:
        svc = self._make_zettel_service([])

        vs = ValidationService()
        vs.output_folder = tmp_path
        config = MagicMock()
        config.__getitem__ = lambda self, k: str(tmp_path) if k == "site_dir" else ""
        files = MagicMock()

        vs.validate(svc, files, config)

        report = tmp_path / "validation.md"
        assert report.exists()
        content = report.read_text()
        assert "Validation Report" in content



"""E2E tests for basic accessibility: ARIA, visually-hidden, focus, headings."""

from conftest import ZETTEL_DEMO, ZETTEL_INSTALL


def test_visually_hidden_elements_exist(page, default_site):
    page.goto(f"{default_site}/{ZETTEL_INSTALL}/")
    sr = page.locator(".visually-hidden")
    assert sr.count() >= 1


def test_settings_button_has_aria_label(page, default_site):
    page.goto(default_site)
    toggle = page.locator('[data-target="#mkdocs_settings_modal"]')
    label = toggle.get_attribute("aria-label")
    assert label is not None
    assert len(label) > 0


def test_search_modal_is_dialog_element(page, default_site):
    page.goto(default_site)
    modal = page.locator("dialog#mkdocs_search_modal")
    assert modal.count() == 1
    assert modal.get_attribute("aria-labelledby") is not None


def test_decorative_icons_are_aria_hidden(page, default_site):
    page.goto(f"{default_site}/{ZETTEL_INSTALL}/")
    # validation badge icon should be aria-hidden
    icons = page.locator('i.fa[aria-hidden="true"]')
    assert icons.count() >= 1


def test_copy_buttons_have_aria_label(page, default_site):
    page.goto(f"{default_site}/{ZETTEL_DEMO}/")
    page.wait_for_selector(".copy-btn")
    btn = page.locator(".copy-btn").first
    assert btn.get_attribute("aria-label") == "Copy code"


def test_focus_ring_visible_on_tab(page, default_site):
    page.goto(default_site)
    page.keyboard.press("Tab")
    focused = page.evaluate("""() => {
        const el = document.activeElement;
        if (!el) return null;
        const style = getComputedStyle(el);
        return {
            outline: style.outline,
            outlineWidth: style.outlineWidth,
            tag: el.tagName
        };
    }""")
    assert focused is not None
    assert focused["tag"] != "BODY"


def test_heading_hierarchy(page, default_site):
    page.goto(f"{default_site}/{ZETTEL_INSTALL}/")
    h1_count = page.locator("h1").count()
    assert h1_count >= 1
    # no h3 without h2
    if page.locator("h3").count() > 0:
        assert page.locator("h2").count() > 0



from pathlib import Path
from unittest.mock import Mock, patch

import pytest

from mkdocs_zettelkasten.plugin.entities.zettel import (
    Zettel,
    ZettelFormatError,
)

VALID_ZETTEL = """---
id: 20240101120000
title: My Zettel
tags:
  - test
date: 2024-01-01
---
# Heading

Some body text with [[20240102120000]] and [link](other.md).
"""

NO_TITLE_ZETTEL = """---
id: 20240101120000
date: 2024-01-01
---
# Alt Title From Heading

Body text.
"""

NO_TITLE_NO_HEADING = """---
id: 20240101120000
date: 2024-01-01
---
Body without heading.
"""

MISSING_ID_ZETTEL = """---
title: No ID Here
---
# Content
"""

UNCLOSED_HEADER = """---
id: 123
title: Unclosed
# Missing closing divider
"""

BOM_CONTENT = (
    "\ufeff---\nid: 20240101120000\ntitle: BOM Test\ndate: 2024-01-01\n---\nBody\n"
)

INVALID_YAML = '---\nkey: "unclosed string\n---\nBody\n'

EXTRA_DIVIDER = """---
id: 20240101120000
title: Extra divider
date: 2024-01-01
---
Body before third divider
---
Content after third divider
"""

HR_WITH_LINKS = """---
id: 20240101120000
title: HR with links
date: 2024-01-01
---
Before the rule [[20240102120000]]
---
After the rule [[20240103120000]]
"""


def _make_zettel(tmp_path: Path, content: str, **kwargs) -> Zettel:
    fp = tmp_path / "test.md"
    fp.write_text(content)
    mock_stat = Mock()
    mock_stat.st_mtime = 1704067200.0  # 2024-01-01
    with (
        patch.object(Path, "stat", return_value=mock_stat),
        patch(
            "mkdocs_zettelkasten.plugin.entities.zettel.GitUtil.is_tracked",
            return_value=False,
        ),
    ):
        return Zettel(fp, str(fp.relative_to(tmp_path)), **kwargs)


class TestZettelInit:
    def test_valid_zettel(self, tmp_path: Path) -> None:
        z = _make_zettel(tmp_path, VALID_ZETTEL)
        assert z.id == 20240101120000
        assert z.title == "My Zettel"
        assert z.last_update_date == "2024-01-01"

    def test_missing_id_raises(self, tmp_path: Path) -> None:
        with pytest.raises(ZettelFormatError, match="Missing zettel ID"):
            _make_zettel(tmp_path, MISSING_ID_ZETTEL)

    def test_unclosed_header_raises(self, tmp_path: Path) -> None:
        with pytest.raises(ZettelFormatError, match="Unclosed YAML header"):
            _make_zettel(tmp_path, UNCLOSED_HEADER)

    def test_invalid_yaml_raises(self, tmp_path: Path) -> None:
        with pytest.raises(ZettelFormatError):
            _make_zettel(tmp_path, INVALID_YAML)

    def test_bom_encoding(self, tmp_path: Path) -> None:
        z = _make_zettel(tmp_path, BOM_CONTENT)
        assert z.id == 20240101120000
        assert z.title == "BOM Test"


class TestTitleFallback:
    def test_title_from_frontmatter(self, tmp_path: Path) -> None:
        z = _make_zettel(tmp_path, VALID_ZETTEL)
        assert z.title == "My Zettel"

    def test_title_from_heading(self, tmp_path: Path) -> None:
        z = _make_zettel(tmp_path, NO_TITLE_ZETTEL)
        assert z.title == "Alt Title From Heading"

    def test_title_from_filename(self, tmp_path: Path) -> None:
        fp = tmp_path / "20240101120000_my_note.md"
        fp.write_text(NO_TITLE_NO_HEADING)
        mock_stat = Mock()
        mock_stat.st_mtime = 1704067200.0
        with (
            patch.object(Path, "stat", return_value=mock_stat),
            patch(
                "mkdocs_zettelkasten.plugin.entities.zettel.GitUtil.is_tracked",
                return_value=False,
            ),
        ):
            z = Zettel(fp, str(fp.relative_to(tmp_path)))
        assert z.title == "My note"


class TestExtractLinks:
    def test_extracts_wiki_links(self, tmp_path: Path) -> None:
        z = _make_zettel(tmp_path, VALID_ZETTEL)
        assert "20240102120000" in z.links

    def test_extracts_md_links(self, tmp_path: Path) -> None:
        z = _make_zettel(tmp_path, VALID_ZETTEL)
        assert "other.md" in z.links

    def test_no_links(self, tmp_path: Path) -> None:
        content = "---\nid: 1\ndate: 2024-01-01\n---\nNo links here.\n"
        z = _make_zettel(tmp_path, content)
        assert z.links == []


class TestExtraDivider:
    def test_links_extracted_after_hr(self, tmp_path: Path) -> None:
        z = _make_zettel(tmp_path, HR_WITH_LINKS)
        assert "20240102120000" in z.links
        assert "20240103120000" in z.links


class TestHashAndEquality:
    def test_hash(self, tmp_path: Path) -> None:
        z = _make_zettel(tmp_path, VALID_ZETTEL)
        assert hash(z) == hash(20240101120000)

    def test_eq_same_id(self, tmp_path: Path) -> None:
        z1 = _make_zettel(tmp_path, VALID_ZETTEL)
        z2 = _make_zettel(tmp_path, VALID_ZETTEL)
        assert z1 == z2

    def test_neq_different_type(self, tmp_path: Path) -> None:
        z = _make_zettel(tmp_path, VALID_ZETTEL)
        assert z != "not a zettel"


class TestParseFrontmatter:
    def test_splits_header_and_body(self) -> None:
        from mkdocs_zettelkasten.plugin.utils.frontmatter import parse_frontmatter

        header, body = parse_frontmatter("---\nid: 1\n---\nbody text\n")
        assert "id: 1" in header
        assert "body text" in body

    def test_no_frontmatter_returns_empty_header(self) -> None:
        from mkdocs_zettelkasten.plugin.utils.frontmatter import parse_frontmatter

        header, body = parse_frontmatter("just text\n")
        assert header == ""
        assert "just text" in body

    def test_unclosed_frontmatter_returns_empty_header(self) -> None:
        from mkdocs_zettelkasten.plugin.utils.frontmatter import parse_frontmatter

        header, _body = parse_frontmatter("---\nid: 1\n")
        assert header == ""


class TestConfigurableKeys:
    def test_custom_id_key(self, tmp_path: Path) -> None:
        content = "---\nzettel_id: 99\ndate: 2024-01-01\n---\n# Title\n"
        z = _make_zettel(tmp_path, content, zettel_config={"id_key": "zettel_id"})
        assert z.id == 99

    def test_custom_date_key(self, tmp_path: Path) -> None:
        content = "---\nid: 1\ncreated: 2025-06-15\n---\n# Title\n"
        z = _make_zettel(tmp_path, content, zettel_config={"date_key": "created"})
        assert z.last_update_date == "2025-06-15"

    def test_custom_last_update_key(self, tmp_path: Path) -> None:
        content = "---\nid: 1\nmodified: 2025-06-15\ndate: 2024-01-01\n---\n# Title\n"
        z = _make_zettel(
            tmp_path, content, zettel_config={"last_update_key": "modified"}
        )
        assert z.last_update_date == "2025-06-15"



from __future__ import annotations

from pathlib import PurePosixPath
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from mkdocs_zettelkasten.plugin.entities.zettel import Zettel

import logging

logger = logging.getLogger(
    __name__.replace("mkdocs_zettelkasten.plugin.", "mkdocs.plugins.zettelkasten.")
)


class ZettelStore:
    """Storage for zettels with efficient lookup capabilities."""

    def __init__(self, zettels: Iterable[Zettel] = ()) -> None:
        self.update(zettels)

    @property
    def zettels(self) -> list[Zettel]:
        """View of stored zettels."""
        return self._zettels

    def _rebuild_indexes(self) -> None:
        """Maintain internal indexes for fast lookups."""
        self._path_index: dict[Path, Zettel] = {z.path: z for z in self._zettels}
        self._suffix_index: dict[tuple[str, ...], Zettel] = {}
        for z in self._zettels:
            stem = z.path.with_suffix("")
            parts = stem.parts
            for i in range(len(parts)):
                key = parts[i:]
                if key not in self._suffix_index:
                    self._suffix_index[key] = z

    def get_by_path(self, path: Path) -> Zettel | None:
        """Retrieve zettel by filesystem path."""
        return self._path_index.get(path)

    def get_by_partial_path(
        self, partial_path: str, file_suffix: str = ".md"
    ) -> Zettel | None:
        """Retrieve zettel by matching tail path segments."""
        partial_parts = PurePosixPath(partial_path.removesuffix(file_suffix)).parts
        return self._suffix_index.get(partial_parts)

    def update(self, zettels: Iterable[Zettel]) -> None:
        """Replace stored zettels with new collection."""
        sorted_zettels = sorted(zettels, key=lambda z: z.id)
        self._zettels = list(dict.fromkeys(sorted_zettels))
        self._rebuild_indexes()
        logger.info("Zettel store updated with %d zettels.", len(self._zettels))



from conftest import ZETTEL_INSTALL
from playwright.sync_api import expect


def test_previews_json_exists(page, preview_site):
    resp = page.request.get(f"{preview_site}/previews.json")
    assert resp.ok
    data = resp.json()
    assert ZETTEL_INSTALL in data
    assert "title" in data[ZETTEL_INSTALL]
    assert "excerpt" in data[ZETTEL_INSTALL]


def test_hover_shows_popover(page, preview_site):
    page.goto(preview_site)
    link = page.locator(f"a[href*='{ZETTEL_INSTALL}']").first
    link.hover()
    popover = page.locator("#zettel-preview-popover")
    popover.wait_for(state="visible", timeout=5000)
    assert popover.is_visible()


def test_popover_contains_title(page, preview_site):
    page.goto(preview_site)
    link = page.locator(f"a[href*='{ZETTEL_INSTALL}']").first
    link.hover()
    popover = page.locator("#zettel-preview-popover")
    popover.wait_for(state="visible", timeout=5000)
    title = popover.locator(".preview-title")
    assert title.text_content() == "Install MkDocs Zettelkasten"


def test_popover_contains_excerpt(page, preview_site):
    page.goto(preview_site)
    link = page.locator(f"a[href*='{ZETTEL_INSTALL}']").first
    link.hover()
    popover = page.locator("#zettel-preview-popover")
    popover.wait_for(state="visible", timeout=5000)
    excerpt = popover.locator(".preview-excerpt")
    assert len(excerpt.text_content()) > 0


def test_popover_hides_on_mouseleave(page, preview_site):
    page.goto(preview_site)
    link = page.locator(f"a[href*='{ZETTEL_INSTALL}']").first
    link.hover()
    popover = page.locator("#zettel-preview-popover")
    popover.wait_for(state="visible", timeout=5000)
    # Move away from the link
    page.locator("h1").first.hover()
    expect(popover).to_have_css("opacity", "0")


def test_no_popover_on_external_link(page, preview_site):
    page.goto(preview_site)
    ext_link = page.locator("a[href*='wikipedia']").first
    ext_link.hover()
    popover = page.locator("#zettel-preview-popover")
    expect(popover).not_to_be_visible()


def test_no_popover_when_disabled(page, default_site):
    page.goto(default_site)
    popover = page.locator("#zettel-preview-popover")
    expect(popover).to_have_count(0)



import pytest

from mkdocs_zettelkasten.plugin.utils.patterns import EMBED_LINK, MD_LINK, WIKI_LINK


class TestWikiLink:
    def test_simple_wiki_link(self) -> None:
        match = WIKI_LINK.search("[[target]]")
        assert match is not None
        assert match.group("url") == "target"
        assert match.group("title") is None

    def test_wiki_link_with_alias(self) -> None:
        match = WIKI_LINK.search("[[target|alias text]]")
        assert match is not None
        assert match.group("url") == "target"
        assert match.group("title") == "alias text"

    def test_multiple_wiki_links(self) -> None:
        matches = list(WIKI_LINK.finditer("see [[one]] and [[two|second]]"))
        assert len(matches) == 2
        assert matches[0].group("url") == "one"
        assert matches[1].group("url") == "two"
        assert matches[1].group("title") == "second"

    def test_wiki_link_with_numeric_id(self) -> None:
        match = WIKI_LINK.search("[[20240101120000]]")
        assert match is not None
        assert match.group("url") == "20240101120000"

    def test_no_wiki_link(self) -> None:
        assert WIKI_LINK.search("no links here") is None

    def test_empty_wiki_link(self) -> None:
        assert WIKI_LINK.search("[[]]") is None


class TestMdLink:
    def test_simple_md_link(self) -> None:
        match = MD_LINK.search("[text](url)")
        assert match is not None
        assert match.group("title") == "text"
        assert match.group("url") == "url"

    def test_md_link_with_path(self) -> None:
        match = MD_LINK.search("[my page](path/to/page.md)")
        assert match is not None
        assert match.group("title") == "my page"
        assert match.group("url") == "path/to/page.md"

    def test_multiple_md_links(self) -> None:
        matches = list(MD_LINK.finditer("[a](1) and [b](2)"))
        assert len(matches) == 2
        assert matches[0].group("url") == "1"
        assert matches[1].group("url") == "2"

    def test_no_md_link(self) -> None:
        assert MD_LINK.search("no links here") is None

    def test_md_link_empty_text(self) -> None:
        match = MD_LINK.search("[](url)")
        assert match is not None
        assert match.group("title") == ""
        assert match.group("url") == "url"


class TestEmbedLink:
    @pytest.mark.parametrize(
        ("text", "url", "section", "title"),
        [
            ("![[20211122194827]]", "20211122194827", None, None),
            ("![[20211122194827#Method 1]]", "20211122194827", "Method 1", None),
            ("![[20211122194827|Custom]]", "20211122194827", None, "Custom"),
            (
                "![[20211122194827#Method 1|Custom]]",
                "20211122194827",
                "Method 1",
                "Custom",
            ),
            ("![[notes/my-note]]", "notes/my-note", None, None),
        ],
    )
    def test_matches_embed_syntax(
        self, text: str, url: str, section: str | None, title: str | None
    ) -> None:
        m = EMBED_LINK.search(text)
        assert m is not None
        assert m.group("url") == url
        assert m.group("section") == section
        assert m.group("title") == title

    def test_does_not_match_regular_wiki_link(self) -> None:
        assert EMBED_LINK.search("[[20211122194827]]") is None

    def test_does_not_match_image(self) -> None:
        assert EMBED_LINK.search("![alt](image.png)") is None



from __future__ import annotations

import datetime
import re
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pathlib import Path

import logging
from zoneinfo import ZoneInfo

import yaml
from yaml.scanner import ScannerError

from mkdocs_zettelkasten.plugin.utils.date_utils import convert_string_to_date
from mkdocs_zettelkasten.plugin.utils.frontmatter import parse_frontmatter
from mkdocs_zettelkasten.plugin.utils.git_utils import GitUtil
from mkdocs_zettelkasten.plugin.utils.patterns import MD_LINK, WIKI_LINK

logger = logging.getLogger(
    __name__.replace("mkdocs_zettelkasten.plugin.", "mkdocs.plugins.zettelkasten.")
)


class ZettelFormatError(ValueError):
    """Exception raised for invalid zettel file format"""


class Zettel:
    def __init__(
        self,
        abs_src_path: Path,
        src_path: str,
        zettel_config: dict[str, Any] | None = None,
    ) -> None:
        self.id: int = 0
        self.title: str = ""
        self.path: Path = abs_src_path
        self.rel_path: str = src_path
        self.backlinks: list[dict[str, str]] = []
        self.links: list[str] = []
        self.last_update_date: str = ""

        cfg = zettel_config or {}
        self._id_key = cfg.get("id_key", "id")
        self._date_key = cfg.get("date_key", "date")
        self._last_update_key = cfg.get("last_update_key", "last_update")
        self._id_format = cfg.get("id_format", r"^\d{14}$")
        self._tz = cfg.get("timezone") or ZoneInfo("UTC")
        self._date_format = cfg.get("date_format", "%Y-%m-%d")

        self._initialize_zettel()

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Zettel) and self.id == other.id

    def _initialize_zettel(self) -> None:
        """Orchestrates the zettel initialization process."""
        try:
            content = self.path.read_text(encoding="utf-8-sig", errors="strict")
        except OSError as err:
            logger.exception("Failed to read file %s", self.path)
            msg = f"File {self.path} read error"
            raise ZettelFormatError(msg) from err

        header_text, body_text = parse_frontmatter(content)
        if not header_text:
            logger.error("Unclosed YAML header in file: %s", self.path)
            msg = "Unclosed YAML header in file"
            raise ZettelFormatError(msg)

        meta = self._parse_metadata(header_text)
        body_lines = body_text.splitlines()
        self._extract_links(body_lines)
        self._set_core_metadata(meta, self._find_alt_title(body_lines))
        logger.debug(
            "Initialized zettel %s (ID: %s, Title: %s)",
            self.rel_path,
            self.id,
            self.title,
        )

    def _parse_metadata(self, header: str) -> dict:
        """Parses YAML metadata from header text."""
        try:
            meta = yaml.safe_load(header) or {}

            if not isinstance(meta, dict):
                logger.error(
                    "Invalid YAML structure in %s: not a dictionary", self.path
                )
                msg = "Invalid YAML structure"
                raise ZettelFormatError(msg)

            logger.debug("Found metadata keys: %s", list(meta.keys()))

        except (ScannerError, AttributeError) as err:
            logger.exception("Failed to parse YAML in %s", self.path)
            msg = f"Invalid YAML in {self.path}"
            raise ZettelFormatError(msg) from err
        else:
            return meta

    def _find_alt_title(self, body: list[str]) -> str:
        """Finds alternative title from markdown heading."""
        for line in body:
            if line.lstrip().startswith("# "):
                return line.strip()[2:]
        return ""

    def _extract_links(self, body: list[str]) -> None:
        """Extracts all wiki and markdown links from body."""
        wiki_links = []
        markdown_links = []

        for line in body:
            wiki_matches = [m.groupdict()["url"] for m in WIKI_LINK.finditer(line)]
            md_matches = [m.groupdict()["url"] for m in MD_LINK.finditer(line)]
            wiki_links.extend(wiki_matches)
            markdown_links.extend(md_matches)

        self.links.extend(wiki_links)
        self.links.extend(markdown_links)

        logger.debug(
            "Extracted %d wiki links and %d markdown links",
            len(wiki_links),
            len(markdown_links),
        )

    def _set_core_metadata(self, meta: dict, alt_title: str) -> None:
        """Sets fundamental metadata fields."""
        if not meta.get(self._id_key):
            logger.error("Missing required ID in zettel: %s", self.path)
            msg = "Missing zettel ID"
            raise ZettelFormatError(msg)

        self.id = meta[self._id_key]

        if meta.get("title"):
            self.title = meta["title"]
        elif alt_title:
            self.title = alt_title
        else:
            self.title = self._generate_filename_title()

        self._determine_last_update_date(meta)

    def _generate_filename_title(self) -> str:
        """Generates title from filename using pattern matching."""
        stem = self.path.stem
        # Strip ID prefix from filename (remove anchors from id_format pattern)
        id_prefix = self._id_format.lstrip("^").rstrip("$")
        clean_stem = re.sub(id_prefix, "", stem)
        return clean_stem.replace("_", " ").replace("-", " ").strip().capitalize()

    def _determine_last_update_date(self, meta: dict) -> None:
        """Determines most recent valid date from multiple sources."""
        candidate_date = self._get_initial_candidate_date(meta)
        revision_date = self._get_revision_date()

        if self._last_update_key in meta:
            final_date = candidate_date
        else:
            final_date = max(candidate_date, revision_date, key=lambda d: d.timestamp())
            logger.debug(
                "Using later date between metadata (%s) and modification (%s) for %s",
                candidate_date.strftime("%Y-%m-%d"),
                revision_date.strftime("%Y-%m-%d"),
                self.rel_path,
            )

        self.last_update_date = final_date.strftime(self._date_format)

    def _get_initial_candidate_date(self, meta: dict) -> datetime.datetime:
        """Gets first valid date from metadata sources."""
        for field in [self._last_update_key, self._date_key]:
            if date := convert_string_to_date(meta.get(field, ""), tz=self._tz):
                return date

        id_date = convert_string_to_date(str(self.id), tz=self._tz)
        if id_date:
            return id_date

        return datetime.datetime.now(tz=self._tz)

    def _get_revision_date(self) -> datetime.datetime:
        """Gets revision date from VCS or filesystem."""
        if GitUtil.is_tracked(str(self.path)):
            git_date = GitUtil().get_revision_date_for_file(str(self.path))
            if git_date is not None:
                return git_date

        return self._get_mtime()

    def _get_mtime(self) -> datetime.datetime:
        """Gets modification time from filesystem."""
        st_mtime = self.path.stat().st_mtime
        return datetime.datetime.fromtimestamp(st_mtime, tz=self._tz)



from unittest.mock import MagicMock

from mkdocs.structure.files import File

from mkdocs_zettelkasten.plugin.adapters.prev_next_page import get_prev_next_page


def _make_file(src_path: str, abs_src_path: str | None = None) -> MagicMock:
    f = MagicMock(spec=File)
    f.src_path = src_path
    f.abs_src_path = abs_src_path or f"/docs/{src_path}"
    f.page = MagicMock()
    return f


def _make_zettel(zettel_id: int, path: str) -> MagicMock:
    z = MagicMock()
    z.id = zettel_id
    z.path = path
    return z


def _make_page(src_path: str, is_zettel: bool = True, zettel_id: int = 1) -> MagicMock:
    page = MagicMock()
    page.file.src_path = src_path
    page.meta = {"is_zettel": is_zettel}
    if is_zettel:
        zettel = MagicMock()
        zettel.id = zettel_id
        page.meta["zettel"] = zettel
    return page


class TestGetPrevNextPage:
    def test_index_page_returns_first_zettel(self) -> None:
        page = _make_page("index.md", is_zettel=False)
        z = _make_zettel(1, "/docs/note.md")
        f = _make_file("note.md", "/docs/note.md")
        files = MagicMock()
        files.__iter__ = lambda self: iter([f])

        _, next_page = get_prev_next_page(page, files, [z])
        assert next_page is not None

    def test_tags_page_returns_none(self) -> None:
        page = _make_page("tags.md", is_zettel=False)
        files = MagicMock()
        files.__iter__ = lambda self: iter([])

        prev, next_p = get_prev_next_page(page, files, [])
        assert prev is None
        assert next_p is None

    def test_non_zettel_returns_none(self) -> None:
        page = _make_page("about.md", is_zettel=False)
        files = MagicMock()
        files.__iter__ = lambda self: iter([])

        prev, next_p = get_prev_next_page(page, files, [])
        assert prev is None
        assert next_p is None

    def test_zettel_navigation(self) -> None:
        z1 = _make_zettel(1, "/docs/a.md")
        z2 = _make_zettel(2, "/docs/b.md")
        z3 = _make_zettel(3, "/docs/c.md")

        f_index = _make_file("index.md")
        f1 = _make_file("a.md", "/docs/a.md")
        f2 = _make_file("b.md", "/docs/b.md")
        f3 = _make_file("c.md", "/docs/c.md")

        files = MagicMock()
        files.__iter__ = lambda self: iter([f_index, f1, f2, f3])

        page = _make_page("b.md", zettel_id=2)

        prev, next_p = get_prev_next_page(page, files, [z1, z2, z3])
        assert prev is f1.page
        assert next_p is f3.page

    def test_first_zettel_has_homepage_as_prev(self) -> None:
        z1 = _make_zettel(1, "/docs/a.md")
        z2 = _make_zettel(2, "/docs/b.md")

        f_index = _make_file("index.md")
        f1 = _make_file("a.md", "/docs/a.md")
        f2 = _make_file("b.md", "/docs/b.md")

        files = MagicMock()
        files.__iter__ = lambda self: iter([f_index, f1, f2])

        page = _make_page("a.md", zettel_id=1)

        prev, next_p = get_prev_next_page(page, files, [z1, z2])
        assert prev is f_index.page
        assert next_p is f2.page

    def test_last_zettel_has_no_next(self) -> None:
        z1 = _make_zettel(1, "/docs/a.md")
        z2 = _make_zettel(2, "/docs/b.md")

        f_index = _make_file("index.md")
        f1 = _make_file("a.md", "/docs/a.md")
        f2 = _make_file("b.md", "/docs/b.md")

        files = MagicMock()
        files.__iter__ = lambda self: iter([f_index, f1, f2])

        page = _make_page("b.md", zettel_id=2)

        prev, next_p = get_prev_next_page(page, files, [z1, z2])
        assert prev is f1.page
        assert next_p is None



#!/usr/bin/env python3
"""WCAG 2.1 contrast-ratio checker for color-scheme CSS files."""

from __future__ import annotations

import re
import sys
from pathlib import Path

# ---------------------------------------------------------------------------
# Named CSS colors used in scheme files
# ---------------------------------------------------------------------------
NAMED_COLORS: dict[str, str] = {
    "black": "#000000",
    "white": "#ffffff",
    "gray": "#808080",
    "grey": "#808080",
    "lightgray": "#d3d3d3",
    "lightgrey": "#d3d3d3",
    "darkgray": "#a9a9a9",
    "darkgrey": "#a9a9a9",
    "silver": "#c0c0c0",
    "red": "#ff0000",
    "green": "#008000",
    "blue": "#0000ff",
    "yellow": "#ffff00",
    "orange": "#ffa500",
}

# ---------------------------------------------------------------------------
# Contrast pairs: (foreground var, background var, required ratio, label)
# ---------------------------------------------------------------------------
CONTRAST_PAIRS: list[tuple[str, str, float, str]] = [
    ("--text-primary", "--bg-zettel", 4.5, "body text"),
    ("--text-link", "--bg-zettel", 4.5, "links"),
    ("--text-header", "--bg-zettel", 3.0, "headings (large)"),
    ("--text-primary", "--bg-accent", 4.5, "text on accent bg"),
    ("--text-code", "--bg-code", 4.5, "code blocks"),
]

# ---------------------------------------------------------------------------
# Color helpers
# ---------------------------------------------------------------------------

_HEX3 = re.compile(r"^#([0-9a-fA-F]{3})$")
_HEX6 = re.compile(r"^#([0-9a-fA-F]{6})$")


def parse_color(raw: str) -> tuple[int, int, int]:
    """Return (R, G, B) 0-255 from a hex string or named color."""
    value = raw.strip().lower()
    if value in NAMED_COLORS:
        value = NAMED_COLORS[value]

    m6 = _HEX6.match(value)
    if m6:
        h = m6.group(1)
        return int(h[0:2], 16), int(h[2:4], 16), int(h[4:6], 16)

    m3 = _HEX3.match(value)
    if m3:
        h = m3.group(1)
        return int(h[0] * 2, 16), int(h[1] * 2, 16), int(h[2] * 2, 16)

    msg = f"cannot parse color: {raw!r}"
    raise ValueError(msg)


def _linearize(c8: int) -> float:
    """sRGB channel (0-255) -> linear value per WCAG 2.1."""
    s = c8 / 255.0
    return s / 12.92 if s <= 0.04045 else ((s + 0.055) / 1.055) ** 2.4


def relative_luminance(r: int, g: int, b: int) -> float:
    """WCAG 2.1 relative luminance (0-1)."""
    return 0.2126 * _linearize(r) + 0.7152 * _linearize(g) + 0.0722 * _linearize(b)


def contrast_ratio(fg: tuple[int, int, int], bg: tuple[int, int, int]) -> float:
    """WCAG contrast ratio (1-21)."""
    l1 = relative_luminance(*fg)
    l2 = relative_luminance(*bg)
    if l1 < l2:
        l1, l2 = l2, l1
    return (l1 + 0.05) / (l2 + 0.05)


# ---------------------------------------------------------------------------
# CSS parsing
# ---------------------------------------------------------------------------

_SELECTOR_RE = re.compile(
    r'\[data-color-scheme="([^"]+)"\]'
    r'(?:\[data-theme="dark"\])?'
)
_VAR_RE = re.compile(r"(--[\w-]+)\s*:\s*([^;]+);")


def parse_scheme_file(path: Path) -> dict[str, dict[str, str]]:
    """Parse a scheme CSS file into {variant: {var: value}}.

    Returns a dict keyed by "<scheme_id>" for light and
    "<scheme_id>__dark" for the dark variant.
    """
    text = path.read_text()
    result: dict[str, dict[str, str]] = {}

    # Split on top-level selectors while keeping the selector text.
    # Strategy: find each selector block via brace matching.
    pos = 0
    while pos < len(text):
        sel_match = _SELECTOR_RE.search(text, pos)
        if not sel_match:
            break
        scheme_id = sel_match.group(1)
        is_dark = '[data-theme="dark"]' in sel_match.group(0)
        key = f"{scheme_id}__dark" if is_dark else scheme_id

        # Find the opening brace after the selector
        brace_start = text.index("{", sel_match.end())
        depth = 1
        i = brace_start + 1
        while i < len(text) and depth > 0:
            if text[i] == "{":
                depth += 1
            elif text[i] == "}":
                depth -= 1
            i += 1
        block = text[brace_start + 1 : i - 1]

        variables: dict[str, str] = {}
        for vm in _VAR_RE.finditer(block):
            variables[vm.group(1)] = vm.group(2).strip()
        result[key] = variables
        pos = i

    return result


# ---------------------------------------------------------------------------
# Checker
# ---------------------------------------------------------------------------


def check_schemes(schemes_dir: Path) -> list[dict]:
    """Run contrast checks on all CSS files.  Returns list of result dicts."""
    results: list[dict] = []
    for css_file in sorted(schemes_dir.glob("*.css")):
        variants = parse_scheme_file(css_file)
        for variant_key, variables in sorted(variants.items()):
            is_dark = variant_key.endswith("__dark")
            scheme_name = variant_key.removesuffix("__dark") if is_dark else variant_key
            mode = "dark" if is_dark else "light"

            for fg_var, bg_var, required, label in CONTRAST_PAIRS:
                fg_raw = variables.get(fg_var)
                bg_raw = variables.get(bg_var)
                if fg_raw is None or bg_raw is None:
                    results.append(
                        {
                            "scheme": scheme_name,
                            "mode": mode,
                            "pair": label,
                            "fg_var": fg_var,
                            "bg_var": bg_var,
                            "ratio": None,
                            "required": required,
                            "pass": False,
                            "note": "missing variable",
                        }
                    )
                    continue

                try:
                    fg = parse_color(fg_raw)
                    bg = parse_color(bg_raw)
                except ValueError as exc:
                    results.append(
                        {
                            "scheme": scheme_name,
                            "mode": mode,
                            "pair": label,
                            "fg_var": fg_var,
                            "bg_var": bg_var,
                            "ratio": None,
                            "required": required,
                            "pass": False,
                            "note": str(exc),
                        }
                    )
                    continue

                ratio = contrast_ratio(fg, bg)
                results.append(
                    {
                        "scheme": scheme_name,
                        "mode": mode,
                        "pair": label,
                        "fg_var": fg_var,
                        "bg_var": bg_var,
                        "ratio": ratio,
                        "required": required,
                        "pass": ratio >= required,
                        "note": "",
                    }
                )

    return results


# ---------------------------------------------------------------------------
# Report
# ---------------------------------------------------------------------------


def _print_report(results: list[dict]) -> int:
    """Print a table and return exit code (0 = all pass, 1 = failures)."""
    failures = [r for r in results if not r["pass"]]

    # Print all results grouped by scheme+mode
    current = ""
    for r in results:
        header = f"{r['scheme']} ({r['mode']})"
        if header != current:
            current = header
            print(f"\n  {header}")
            print(f"  {'─' * 60}")

        status = "PASS" if r["pass"] else "FAIL"
        ratio_str = f"{r['ratio']:.2f}" if r["ratio"] is not None else "N/A"
        note = f"  ({r['note']})" if r["note"] else ""
        print(
            f"    [{status}] {r['pair']:<22} "
            f"{ratio_str:>6}:1  (need {r['required']:.1f}:1){note}"
        )

    # Summary
    total = len(results)
    passed = total - len(failures)
    print(f"\n  {passed}/{total} checks passed")

    if failures:
        print(f"\n  {len(failures)} failure(s):")
        for r in failures:
            ratio_str = f"{r['ratio']:.2f}" if r["ratio"] is not None else "N/A"
            note = f" ({r['note']})" if r["note"] else ""
            print(
                f"    - {r['scheme']} {r['mode']}: {r['pair']} "
                f"= {ratio_str}:1 < {r['required']:.1f}:1{note}"
            )
        return 1

    print("\n  All checks passed.")
    return 0


def main(schemes_dir: Path | None = None) -> int:
    if schemes_dir is None:
        schemes_dir = (
            Path(__file__).resolve().parents[2]
            / "mkdocs_zettelkasten"
            / "themes"
            / "zettelkasten"
            / "css"
            / "schemes"
        )
    if not schemes_dir.is_dir():
        print(f"error: schemes dir not found: {schemes_dir}", file=sys.stderr)
        return 1

    results = check_schemes(schemes_dir)
    return _print_report(results)


if __name__ == "__main__":
    raise SystemExit(main())



"""E2E tests for navigation: prev/next links, search modal, keyboard shortcuts."""

from conftest import ZETTEL_DEMO, ZETTEL_INSTALL, ZETTEL_PUBLISH

# 20211122195311 (Run a demo) has both prev and next links
TARGET = f"/{ZETTEL_DEMO}/"


def test_prev_link_exists_and_navigates(page, default_site):
    page.goto(f"{default_site}{TARGET}")
    prev_link = page.locator('a[rel="prev"]')
    assert prev_link.count() == 1
    href = prev_link.get_attribute("href")
    assert href is not None
    assert ZETTEL_INSTALL in href
    prev_link.click()
    page.wait_for_url(f"**/{ZETTEL_INSTALL}/**")


def test_next_link_exists_and_navigates(page, default_site):
    page.goto(f"{default_site}{TARGET}")
    next_link = page.locator('a[rel="next"]')
    assert next_link.count() == 1
    href = next_link.get_attribute("href")
    assert href is not None
    assert ZETTEL_PUBLISH in href
    next_link.click()
    page.wait_for_url(f"**/{ZETTEL_PUBLISH}/**")


def test_search_modal_opens_on_click(page, default_site):
    page.goto(f"{default_site}{TARGET}")
    page.click('a[data-target="#mkdocs_search_modal"]')
    modal = page.locator("#mkdocs_search_modal")
    modal.wait_for(state="visible")
    assert page.locator("#mkdocs-search-query").is_visible()


def test_search_modal_opens_with_s_key(page, default_site):
    page.goto(f"{default_site}{TARGET}")
    page.keyboard.press("s")
    modal = page.locator("#mkdocs_search_modal")
    modal.wait_for(state="visible")
    assert page.locator("#mkdocs-search-query").is_visible()


def test_keyboard_shortcut_n_navigates_next(page, default_site):
    page.goto(f"{default_site}{TARGET}")
    page.keyboard.press("n")
    page.wait_for_url(f"**/{ZETTEL_PUBLISH}/**")


def test_keyboard_shortcut_p_navigates_prev(page, default_site):
    page.goto(f"{default_site}{TARGET}")
    page.keyboard.press("p")
    page.wait_for_url(f"**/{ZETTEL_INSTALL}/**")
